#ifndef _LIBSM
#define _LIBSM

#ifdef __cplusplus
extern "C" {
#endif

#include <stdarg.h>
#include <sys/types.h>
#include <node.h>

/* Code retour des fonctions constituant l'API */

typedef long SMT_Status;

#define SMS_OK          NDS_OK      /* La fonction s'est correctement excute et a produit un rsultat */
#define SMS_KO          NDS_KO      /* La fonction s'est correctement excute mais n'a pas produit de rsultat */
#define SMS_NO_WAIT     2			/* Opration sur smaphore en mode IPC_NOWAIT qui n'a pas pu aboutir */

#define SMS_YES         SMS_OK      /* Rsultat boolen positif */
#define SMS_NO          SMS_KO      /* Rsultat boolen ngatif */

#define SMS_ERRMEM      NDS_ERRMEM  /* Problme d'allocation mmoire */
#define SMS_ERRAPI      NDS_ERRAPI  /* Utilisation incorrecte des API */
#define SMS_ERRSHM      -3          /* Problme relatif aux segments de mmoire partage */
#define SMS_ERRSEM      -4          /* Problme relatif  l'utilisation des smaphores */
#define SMS_ERRSIG      -5          /* Opration sur smaphore interrompue par un signal */

#define SM_ERROR(s)     ((s) < 0)     /* Tous les codes retour ngatifs correspondent  des erreurs */

typedef int SMT_Flags;

#define SMD_UNDEF       0

/* Flags d'ouverture d'un heap ou d'une instance de la librairie */

#define SMD_CREATE      	0x01    /* Cration */
#define SMD_OPEN        	0x02    /* Ouverture simple */
#define SMD_DESTROY     	0x04    /* Destruction */
#define SMD_CLOSE       	0x08    /* Fermeture simple */

/* NB : pour l'ouverture de la librairie, les valeurs SMD_OPEN et SMD_CREATE sont mutuellement exclusives */

/* Flags de verrouillage */

#define SMD_READ        	0x04    /* verrou partag */
#define SMD_WRITE       	0x08    /* verrou exclusif */
#define SMD_NO_LOCK       	0x00    /* aucun verrou */

#define SMD_LOCK_MSK(a)   	((a) & (SMD_READ|SMD_WRITE))

/* Mode de debug sur l'ouverture de la librairie */

#define SMD_DEBUG_NONE  	0x00    /* pour n'afficher aucun message gnr par les diverses librairies */
#define SMD_DEBUG       	0x10    /* pour afficher les messages gnrs par la librairie */
#define SMD_DEBUG_ALL   	0x20    /* pour afficher les messages gnrs par toutes les librairies sous-jacentes */

#define SMD_DEBUG_MSK(a)    ((a) & (SMD_DEBUG|SMD_DEBUG_ALL))

/* Diffrentes types de configuration d'un heap */

typedef int SMT_Config;

#define SMD_SEGMENT_SIZE    	1	/* Dfinir la taille des segments */
#define SMD_HEAP_LIMIT      	2	/* Dfinir la taille maximale */
#define SMD_AUTO_COMPRESS      	3	/* Dfinir la taille d'un FCR  partir de laquelle
									   la compression du heap est autmatiquement active */

#define SMD_DEFAULT_COMPRESS    1000  /* si + de 1000 chunks libres, alors compression du heap */

/* Diffrentes valeurs de configuration */

#define SMD_UNLIMITED           0
#define SMD_NO_AUTO_COMPRESS    0

/* Nom du heap systme */

#define HEAP_SYSTEM  "system"

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE 1
#endif

char SM_Error_Msg [512];

#ifndef SM_MODE
#define SM_MODE 0
#endif

#if SM_MODE == 1 /* API sans vrification des arguments */

#define SM_Library_Open             SM_Library_Open_I
#define SM_Library_Instance_Get     SM_Library_Instance_Get_I
#define SM_Library_Context_Set      SM_Library_Context_Set_I
#define SM_Library_Context_Get      SM_Library_Context_Get_I
#define SM_Library_Close            SM_Library_Close_I
#define SM_Library_Dump             SM_Library_Dump_I
#define SM_Library_Unlock           SM_Library_Unlock_I
#define SM_Library_Stderr_Set       SM_Library_Stderr_Set_I
#define SM_Heap_Exist               SM_Heap_Exist_I
#define SM_Heap_Open                SM_Heap_Open_I
#define SM_Heap_IsOpen              SM_Heap_IsOpen_I
#define SM_Heap_Close               SM_Heap_Close_I
#define SM_Heap_End                 SM_Heap_End_I
#define SM_Heap_Compress            SM_Heap_Compress_I
#define SM_Heap_Config              SM_Heap_Config_I
#define SM_Heap_Check               SM_Heap_Check_I
#define SM_Heap_Lock                SM_Heap_Lock_I
#define SM_Heap_Unlock              SM_Heap_Unlock_I
#define SM_Chunk_Alloc              SM_Chunk_Alloc_I
#define SM_Chunk_Free               SM_Chunk_Free_I

#else /* API avec vrification des arguments */

#define SM_Library_Open             SM_Library_Open_C
#define SM_Library_Instance_Get     SM_Library_Instance_Get_C
#define SM_Library_Context_Set      SM_Library_Context_Set_C
#define SM_Library_Context_Get      SM_Library_Context_Get_C
#define SM_Library_Close            SM_Library_Close_C
#define SM_Library_Dump             SM_Library_Dump_C
#define SM_Library_Unlock           SM_Library_Unlock_C
#define SM_Library_Stderr_Set       SM_Library_Stderr_Set_C
#define SM_Heap_Exist               SM_Heap_Exist_C
#define SM_Heap_Open                SM_Heap_Open_C
#define SM_Heap_IsOpen              SM_Heap_IsOpen_C
#define SM_Heap_Close               SM_Heap_Close_C
#define SM_Heap_End                 SM_Heap_End_C
#define SM_Heap_Compress            SM_Heap_Compress_C
#define SM_Heap_Config              SM_Heap_Config_C
#define SM_Heap_Check               SM_Heap_Check_C
#define SM_Heap_Lock                SM_Heap_Lock_C
#define SM_Heap_Unlock              SM_Heap_Unlock_C
#define SM_Chunk_Alloc              SM_Chunk_Alloc_C
#define SM_Chunk_Free               SM_Chunk_Free_C

#endif

/*------------------------------------------------------------------------------*/
/* Structure de la base de heaps                                                */
/*------------------------------------------------------------------------------*/

typedef struct {
    int         SysMemID;     /* ID de la 1re zone de mmoire partage de la base */
    int         DataMemID;    /* ID de la 2me zone de mmoire partage de la base */
    size_t      Size;         /* Taille de la zone de mmoire partage */
    int         SemID;        /* ID du smaphore pour la gestion des verrous */
    NDT_Root *  MHR;          /* Memory Heap Root : racine de la liste de heap */
    pid_t       Creator;      /* ID du processus crateur de la base */
    pid_t       Writer;       /* ID du dernier processus ayant accd en criture  la base */
    void *      Free;         /* Pointeur sur la premire zone libre de la base */
    void *      Attach;       /* Adresse du dernier attachement */
    size_t      Segment_Size; /* Taille par dfaut des segments qui composeront les heaps */
} SMT_Base;

/* Rfrence sur la base de heaps */

SMT_Base * SM_Base;

/*------------------------------------------------------------------------------*/
/* Structure d'un MHH (Memory Heap Header)                                      */
/* Rappel : un MHH est une valeur attache  un MHN (Memory Heap Node)          */
/*------------------------------------------------------------------------------*/

typedef struct {
    char        Name [256];     /* Nom du heap */
    int         SemID;          /* ID du smaphore pour la gestion des verrous */
    NDT_Root *  DSR;            /* Data Segment Root */
    NDT_Root *  ACR;            /* Allocated Chunks Root */
    NDT_Root *  FCR;            /* Free Chunks Root */
    pid_t       Writer;         /* ID du processus ayant accd en dernier au MHH en criture */
    int         State;          /* Etat d'un heap (valide, non valid ou corrompu) */
    size_t      Segment_Size;   /* Taille des segments de mmoire composant le heap */
    size_t      Limit_Size;     /* Taille limite du heap (par dfaut : pas de limite) */
    int         Auto_Compress;  /* Nombre de chunks libres  partir duquel le heap est automatiquement compress */
} SMT_MHH;

/* Heap ouvert */

typedef struct {
    char    *   Name;
    SMT_MHH *   MHH;
    SMT_Flags   Lock_Mode;  /* Mode dans lequel le heap est verrouill */
    int         Nb_Seg;     /* Nombre de segments du heap lors de son ouverture */
} SMT_Heap;

/*------------------------------------------------------------------------------*/
/* Structure d'un DSH (Data Segment Header)                                     */
/* Rappel : un DSH est une valeur attache  un DSN (noeud de DSR)              */
/*------------------------------------------------------------------------------*/

typedef struct {
    int         MemID;      /* ID de la zone de mmoire partage */
    size_t      Size;       /* Taille de la zone de mmoire partage */
    void *      Start;      /* Adresse de dbut de la zone de mmoire partage */
} SMT_DSH;

/*------------------------------------------------------------------------------*/
/* Structure d'un chunk                                                         */
/* Rappel : un chunk est la valeur attache  un noeud de ACR ou FCR            */
/*------------------------------------------------------------------------------*/

typedef struct {
    size_t      Size;       /* Taille alloue au chunk */
    void *      Data;       /* Adresse de la zone de donnes du chunk */
} SMT_Chunk;

/*------------------------------------------------------------------------------*/
/* Ouverture d'une instance de la librairie                                     */
/*------------------------------------------------------------------------------*/
/* (I) Instance : numro d'instance de la librairie                             */
/* (I) Context  : nom du nouveau contexte                                       */
/* (I) Flags    : indicateur cration/ouverture + mode d'affichage des erreurs  */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Library_Open_I ( int Instance, const char * Context, SMT_Flags Flags );
SMT_Status SM_Library_Open_C ( int Instance, const char * Context, SMT_Flags Flags );

/*------------------------------------------------------------------------------*/
/* Rcupration du numro de l'instance utilise                                */
/*------------------------------------------------------------------------------*/
/* (O) Instance : adresse du numro de l'instance utilise                      */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Library_Instance_Get_I ( int * Instance);
SMT_Status SM_Library_Instance_Get_C ( int * Instance);

/*------------------------------------------------------------------------------*/
/* Changement de contexte d'utilisation de la librairie                         */
/*------------------------------------------------------------------------------*/
/* (I) Context : nom du nouveau contexte                                        */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Library_Context_Set_I ( const char * Context );
SMT_Status SM_Library_Context_Set_C ( const char * Context );

/*------------------------------------------------------------------------------*/
/* Rcupration du nom du contexte utilis                                      */
/*------------------------------------------------------------------------------*/
/* (O) Context : adresse du nom du contexte utilis                             */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Library_Context_Get_I ( char ** Context );
SMT_Status SM_Library_Context_Get_C ( char ** Context );

/*------------------------------------------------------------------------------*/
/* Fermeture de l'instance de la librairie                                      */
/*------------------------------------------------------------------------------*/
/* (I) Flags : mode de fermeture (destruction ou fermeture simple)              */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Library_Close_I ( SMT_Flags Flags );
SMT_Status SM_Library_Close_C ( SMT_Flags Flags );

/*------------------------------------------------------------------------------*/
/*  Affichage des informations de la base de mmoires partages                 */
/*------------------------------------------------------------------------------*/
/* (I) Out : pointeur sur le flux de sortie                                     */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Library_Dump_I ( FILE * Out );
SMT_Status SM_Library_Dump_C ( FILE * Out );

/*------------------------------------------------------------------------------*/
/*  Libration de tous les verrous (base, heap)                                 */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Library_Unlock_I ( void );
SMT_Status SM_Library_Unlock_C ( void );

/*------------------------------------------------------------------------------*/
/* Dfinition de la sortie standard des messages d'erreur de la librairie       */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Library_Stderr_Set_I ( FILE * Out );
SMT_Status SM_Library_Stderr_Set_C ( FILE * Out );

/*------------------------------------------------------------------------------*/
/* Test d'existence d'un heap                                                   */
/*------------------------------------------------------------------------------*/
/* (I) Heap_Name    : nom du heap                                               */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Heap_Exist_I ( const char * Heap_Name );
SMT_Status SM_Heap_Exist_C ( const char * Heap_Name );

/*------------------------------------------------------------------------------*/
/* Ouverture/cration d'un heap                                                 */
/*------------------------------------------------------------------------------*/
/* (I) Heap_Name : nom du heap                                                  */
/* (O) Heap      : pointeur sur le heap ouvert / cr                           */
/* (I) Seg_Size  : taille des segments du heap                                  */
/* (I) Flags     : mode d'ouverture du heap                                     */
/* (O) Locked    : verrou effectif (TRUE ou FALSE)                              */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Heap_Open_I ( const char * Heap_Name, SMT_Heap ** Heap, size_t Seg_Size, SMT_Flags Flags, int * Locked );
SMT_Status SM_Heap_Open_C ( const char * Heap_Name, SMT_Heap ** Heap, size_t Seg_Size, SMT_Flags Flags, int * Locked );

/*------------------------------------------------------------------------------*/
/* Teste si un heap a dj t ouvert par le processus courant                  */
/*------------------------------------------------------------------------------*/
/* (I) Heap_Name : nom du heap                                                  */
/* (O) Heap      : adresse du pointeur sur le heap ouvert                       */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Heap_IsOpen_I ( const char * Heap_Name, SMT_Heap ** Heap );
SMT_Status SM_Heap_IsOpen_C ( const char * Heap_Name, SMT_Heap ** Heap );

/*------------------------------------------------------------------------------*/
/* Fermeture d'un heap                                                          */
/*------------------------------------------------------------------------------*/
/* (I) Heap   : pointeur sur un heap ouvert                                     */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Heap_Close_I ( SMT_Heap * Heap );
SMT_Status SM_Heap_Close_C ( SMT_Heap * Heap );

/*------------------------------------------------------------------------------*/
/* Destruction d'un heap                                                        */
/*------------------------------------------------------------------------------*/
/* (I) Heap_Name    : nom du heap                                               */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Heap_End_I ( const char * Heap_Name );
SMT_Status SM_Heap_End_C ( const char * Heap_Name );

/*------------------------------------------------------------------------------*/
/* Compression d'un heap                                                        */
/*------------------------------------------------------------------------------*/
/* (I) Heap     : pointeur sur un heap ouvert                                   */
/* (O) Compress : pointeur sur la taille mmoire gagne                         */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Heap_Compress_I ( SMT_Heap * Heap, size_t * Compress );
SMT_Status SM_Heap_Compress_C ( SMT_Heap * Heap, size_t * Compress );

/*------------------------------------------------------------------------------*/
/* Configuration d'un heap                                                      */
/*------------------------------------------------------------------------------*/
/* (I) Heap  : pointeur sur un heap ouvert                                      */
/* (I) Tag   : type de configuration                                            */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Heap_Config_I ( SMT_Heap * Heap, SMT_Config Tag, ... );
SMT_Status SM_Heap_Config_C ( SMT_Heap * Heap, SMT_Config Tag, ... );

/*------------------------------------------------------------------------------*/
/* Vrification/correction des structures d'un heap :                           */
/*------------------------------------------------------------------------------*/
/* (I) Heap         : pointeur sur un heap ouvert                               */
/* (O) Nb_Detected  : pointeur sur le nombre d'erreurs dtectes                */
/* (O) Nb_Corrected : pointeur sur le nombre d'erreurs corriges                */
/* (I) Out          : pointeur sur le flux de sortie du rapport                 */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Heap_Check_I ( SMT_Heap * Heap, int * Nb_Detected, int * Nb_Corrected, FILE * Out);
SMT_Status SM_Heap_Check_C ( SMT_Heap * Heap, int * Nb_Detected, int * Nb_Corrected, FILE * Out);

/*------------------------------------------------------------------------------*/
/* Pose d'un verrou sur un heap                                                 */
/*------------------------------------------------------------------------------*/
/* (I) Heap      : pointeur sur un heap ouvert                                  */
/* (I) Flags     : mode de verrouillage (SMD_READ ou SMD_WRITE)                 */
/* (O) Locked    : verrouillage effectu (TRUE ou FALSE)                        */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Heap_Lock_I ( SMT_Heap * Heap, SMT_Flags Flags, int * Locked );
SMT_Status SM_Heap_Lock_C ( SMT_Heap * Heap, SMT_Flags Flags, int * Locked );

/*------------------------------------------------------------------------------*/
/* Libration d'un verrou sur un heap                                           */
/*------------------------------------------------------------------------------*/
/* (I) Heap   : pointeur sur un heap ouvert                                     */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Heap_Unlock_I ( SMT_Heap * Heap );
SMT_Status SM_Heap_Unlock_C ( SMT_Heap * Heap );

/*------------------------------------------------------------------------------*/
/* Allocation d'un chunk dans un heap                                           */
/*------------------------------------------------------------------------------*/
/* (I) Heap      : pointeur sur un heap ouvert                                  */
/* (I) Size      : taille du chunk                                              */
/* (O) Ptr       : pointeur sur la zone de donnes alloue                      */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Chunk_Alloc_I ( SMT_Heap * Heap, size_t Alloc_Size, void ** Ptr );
SMT_Status SM_Chunk_Alloc_C ( SMT_Heap * Heap, size_t Alloc_Size, void ** Ptr );

/*------------------------------------------------------------------------------*/
/* Dsallocation d'un chunk                                                     */
/*------------------------------------------------------------------------------*/
/* (I) Heap   : pointeur sur un heap ouvert                                     */
/* (I) Ptr    : adresse de la zone de donnes du chunk  dsallouer             */
/*------------------------------------------------------------------------------*/
SMT_Status SM_Chunk_Free_I ( SMT_Heap * Heap, void * Ptr );
SMT_Status SM_Chunk_Free_C ( SMT_Heap * Heap, void * Ptr );

#ifdef __cplusplus
}
#endif

#endif
